/*

	Predator.cpp
	
	© 2001, Stephan Aßmus, All Rights Reserved.
	
	code stolen from EffecTV by FUKUCHI Kentarou
	
	based on HorizontalFlip.cpp sample code provided by Georges-Edouard Berenger

*/

#include "Predator.h"
#include "utils.h"

#include <Bitmap.h>
#include <SupportDefs.h>
#include <InterfaceDefs.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

status_t GetVideoPlugins(TList<VideoPluginType> & plugin_list, VideoPluginsHost * host, image_id id)
{
	if (host->APIVersion() < VideoPluginsHost::VAPI_MINIMAL_API_VERSION)
		return VAPI_HOST_TOO_OLD;

	plugin_list.AddItem(new PredatorPluginType(host, id));
	return B_OK;
}

PredatorPluginType::PredatorPluginType(VideoPluginsHost * host, image_id id)
				: VideoPluginType(host, id)
{
}

PredatorPluginType::~PredatorPluginType()
{
}

const char* PredatorPluginType::ID()
{
	return "Predator";
}

uint32 PredatorPluginType::Flags()
{
	return 0;
}

uint32 PredatorPluginType::Version()
{
	return 001;
}

const char* PredatorPluginType::Name()
{
	return "Predator";
}

const char* PredatorPluginType::Author()
{
	return "stippi";
}

const char* PredatorPluginType::AboutString()
{
	return "Predator by stippi, original code by FUKUCHI Kentarou";
}

status_t PredatorPluginType::About()
{
	return B_OK;
}


static const vertice_properties support[] =
{
	{	B_RGB32,
		B_RGB32,
		1.0,
//		VAPI_PROCESS_IN_PLACE,
		VAPI_PROCESS_TO_DISTINCT,
	},
};

int32 PredatorPluginType::GetVerticesProperties(const vertice_properties ** ap)
{
	*ap = support;
	return sizeof(support) / sizeof(vertice_properties);
}

VideoPluginEngine * PredatorPluginType::InstanciateVideoPluginEngine()
{
	return new PredatorEngine(this);
}

//-------------------------------------------------------------------------

PredatorEngine::PredatorEngine(VideoPluginType * plugin)
			: VideoPluginEngine(plugin),
			bgimage(NULL),
			initialized(false)
{
		yuvTableInit();
}

PredatorEngine::~PredatorEngine()
{
	if (bgimage)
		delete[] bgimage;
}

status_t PredatorEngine::SetConfig(BMessage & config)
{
	return B_OK;
}

status_t PredatorEngine::GetConfig(BMessage & config)
{
	return B_OK;
}

BView* PredatorEngine::ConfigureView()
{
	return NULL;
}

status_t PredatorEngine::UseFormat(media_raw_video_format & format, int which_work_mode)
{
	initialized = false;

	if (bgimage)
		delete[] bgimage;

	width = format.display.line_width;
	height = format.display.line_count;
	bounds.Set(0.0, 0.0, (float)width - 1, (float)height - 1);
	space = format.display.format;

	if (space != B_RGB32)
		return B_BAD_VALUE;

	bgimage = new unsigned int[width * height * 8];
	if (!bgimage)
		return B_NO_MEMORY;

 	bgvalue = (unsigned char *)bgimage + width * height * 4;
	buffer = (unsigned char *)bgvalue + width * height;

	initialized = true;
	bg_taken = false;
	pics_taken = 0;
	return B_OK;
}

status_t PredatorEngine::ApplyEffect(BBitmap * frame_in,
								   BBitmap * frame_out,
								   int64 frame_count,
								   bool skipped_frames)
{
	if (!initialized)
		return B_NO_INIT;
		
	if (!frame_in || !frame_out)
		return B_BAD_VALUE;

	if (frame_in->Bounds() != bounds)
		return B_BAD_VALUE;
	
	if (space != frame_out->ColorSpace())
		return B_BAD_VALUE;
	
	key_info keyInfo;
	if (get_key_info(&keyInfo) == B_OK)
	{
		if (keyInfo.key_states[0x5e>>3] & (1 << (7 - (0x5e%8))))	// 0x5e = keycode for the space bar
		{ 
			bg_taken = false;
			pics_taken = 0;
		}
	}

	if (!bg_taken)
	{
		TakeBackground(frame_in);
		return B_OK;
	}

	int i, x, y;
	unsigned int v;

	// convert to gray!
	int32 pixels = width * height;
	unsigned char *src = (unsigned char *)frame_in->Bits();
	unsigned int *dest = (unsigned int *)frame_out->Bits();

	for(i = 0; i < pixels; i++)
	{
		v = ((int)(*src++ + *src++ + *src++) / 3 - (int)bgvalue[i] + MAGIC_THRESHOLD) & 0x7fffffff;
		buffer[i] = (v - MAGIC_THRESHOLD * 2) >> 31;
		src++;
	}

	i = width + 1;

	for(y = 1; y < height - 1; y++)
	{
		for(x = 1; x < width - 1; x++)
		{
			if (buffer[i - width]
				+ buffer[i - 1] + buffer[i] + buffer[i + 1]
				+ buffer[i + width] > 0)
			{
				dest[i] = bgimage[i];
			}
			else
			{
				dest[i] = bgimage[i + 4] & 0xfcfcfc;
			}
			i++;
		}
		i+=2;
	}
	return B_OK;
}

void PredatorEngine::TakeBackground(BBitmap *src_bitmap)
{
	if (bg_taken)
		return;

	int i;
	unsigned int *src = (unsigned int *)src_bitmap->Bits();
	unsigned int *dest = (unsigned int *)bgvalue;

	// grabs 4 frames and composites them to get a quality background image
	switch (pics_taken)
	{
		case 0:
		{
			// step 1: grab frame-1 to buffer-1
			memcpy(bgimage, src, width * height * 4);
			pics_taken++;
			break;
		}
		case 1:
		{
			// step 2: add frame-2 to buffer-1
			for(i = 0; i < width * height; i++)
				bgimage[i] = (src[i] & bgimage[i]) + (((src[i] ^ bgimage[i]) & 0xfefefe) >> 1);
			pics_taken++;
			break;
		}
		case 2:
		{
			/* step 3: grab frame-3 to buffer-2 */
			memcpy(dest, src, width * height * 4);
			pics_taken++;
			break;
		}
		case 3:
		{
			// step 4: add frame-4 to buffer-2
			for(i = 0; i < width * width; i++)
				dest[i] = (src[i] & dest[i]) + (((src[i] ^ dest[i]) & 0xfefefe) >> 1);
			pics_taken++;
			break;
		}
		case 4:
		{
			// step 5: add buffer-3 to buffer-1
			for(i = 0; i < width * height; i++)
			{
				bgimage[i] = (bgimage[i] & dest[i]) + (((bgimage[i] ^ dest[i]) & 0xfefefe) >> 1);
				bgvalue[i] = RGBtoY(bgimage[i]);
			}
			for ( i = 0; i < 2; i++)
			{
				dest = (unsigned int *)src_bitmap->Bits();
				memcpy(dest, bgimage, width * height * 4);
			}
			bg_taken = true;
			pics_taken++;
			break;
		}
		default:
			return;
	}
}
